package ru.dbotthepony.kstarbound.world.entities

import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.Vector2d
import ru.dbotthepony.kstarbound.math.lerp
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.World
import kotlin.math.absoluteValue

enum class CollisionResolution {
	STOP,
	BOUNCE,
	PUSH,
	SLIDE,
}

/**
 * Определяет из себя сущность в мире, которая имеет позицию, скорость и коробку столкновений
 */
open class Entity(val world: World<*>) {
	var chunk: Chunk? = null
		protected set

	val worldaabb get() = aabb + pos
	var pos = Vector2d()
	var velocity = Vector2d()

	/**
	 * Касается ли сущность земли
	 *
	 * Данный флаг выставляется при обработке скорости, если данный флаг не будет выставлен
	 * правильно, то сущность будет иметь очень плохое движение в стороны
	 *
	 * Так же от него зависит то, может ли сущность двигаться, если она не парит
	 *
	 * Если сущность касается земли, то на неё не действует гравитация
	 */
	var onGround = false
		protected set

	// наследуемые свойства
	open val aabb = AABB.rectangle(Vector2d.ZERO, 0.9, 0.9)
	open val affectedByGravity = true
	open val collisionResolution = CollisionResolution.STOP

	protected open fun onTouchGround(velocity: Vector2d, normal: Vector2d) {

	}

	open fun propagateVelocity(delta: Double) {
		if (velocity.length == 0.0)
			return

		var deltaMovement = velocity * delta

		var potentialAABB = worldaabb + deltaMovement
		var combined = worldaabb.combine(potentialAABB)
		val collected = world.collect((combined).encasingChunkPosAABB()).map { it.chunk.foreground.collisionLayers() }

		if (collected.isNotEmpty()) {
			var newOnGround = false

			for (iteration in 0 .. 100) {
				var collided = false

				for (aabbList in collected) {
					for (aabb in aabbList) {
						if (!newOnGround && iteration == 0) {
							if (worldaabb.sweep(aabb, world.gravity * delta).collisionTime < 1.0) {
								newOnGround = true
							}
						}

						// залез в блоки?
						if (potentialAABB.intersectWeak(aabb)) {
							val push = worldaabb.pushOutFrom(aabb)

							if (push.length > 0.0) {
								velocity -= push * delta * 100.0
								deltaMovement = velocity * delta
								potentialAABB = worldaabb + deltaMovement
								combined = worldaabb.combine(potentialAABB)

								onGround = true
								collided = true
								continue
							}
						}

						// ранний тест (отсечение заведомо не пересекаемой геометрии)
						if (!aabb.intersect(combined)) {
							continue
						}

						// обычный тест коллизии
						val (normal, collisionTime) = worldaabb.sweep(aabb, deltaMovement)

						if (collisionTime != 1.0) {
							val remainingTime = 1.0 - collisionTime
							val oldVelocity = velocity

							when (collisionResolution) {
								CollisionResolution.STOP -> {
									velocity *= remainingTime
								}

								CollisionResolution.PUSH -> {
									var dot = deltaMovement.invDotProduct(normal)
									val magnitude = deltaMovement.length * remainingTime

									if (dot > 0.0) {
										dot = 1.0
									} else {
										dot = -1.0
									}

									velocity = Vector2d(dot * normal.y * magnitude, dot * normal.x * magnitude) / delta
								}

								CollisionResolution.SLIDE -> {
									val dot = deltaMovement.invDotProduct(normal) * remainingTime
									velocity = Vector2d(dot * normal.y, dot * normal.x) / delta
								}

								CollisionResolution.BOUNCE -> {
									velocity *= remainingTime

									if (normal.x.absoluteValue > 0.00001 && normal.y.absoluteValue > 0.00001) {
										velocity *= Vector2d.INVERT_XY
									} else if (normal.x.absoluteValue > 0.00001) {
										velocity *= Vector2d.INVERT_X
									} else if (normal.y.absoluteValue > 0.00001) {
										velocity *= Vector2d.INVERT_Y
									}
								}
							}

							collided = true

							if (!newOnGround) {
								newOnGround = normal.dotProduct(world.gravity) <= -0.98
							}

							deltaMovement = velocity * delta
							potentialAABB = worldaabb + deltaMovement
							onTouchGround(oldVelocity, normal)
						}
					}
				}

				if (!collided) {
					//println("Resolved collision on $iteration")
					break
				}
			}

			onGround = newOnGround
			//println(newOnGround)
		} else {
			onGround = false
		}

		pos += velocity * delta
	}

	open fun moveAndCollide(delta: Double) {
		if (!onGround && affectedByGravity)
			velocity += world.gravity * delta
		else if (affectedByGravity)
			velocity *= Vector2d(lerp(delta, 1.0, 0.01), 1.0)

		propagateVelocity(delta)
	}

	companion object {
		private val LOGGER = LogManager.getLogger(Entity::class.java)
	}
}
